/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.text.DecimalFormat;

import javax.swing.JLabel;

import schmitzm.swing.JPanel;

/**
 * Diese Klasse stellt einen Massstab-Balken ({@link ScalePanel}) und die
 * dazugehoerende Aufloesung in Metern pro Bildschirm-Pixel dar.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ScalePane extends JPanel{
  private static final DecimalFormat numFormat = new DecimalFormat("###,###,##0.00");

  /** Massstab-Balken */
  protected ScalePanel scalePanel  = null;
  /** Label in dem die Aufloesung in Meter pro Pixel angezeigt wird. */
  protected JLabel     scaleLabel  = new JLabel("");

  /**
   * Erzeugt eine Mass-Stab.
   */
  public ScalePane() {
    super();
    this.scalePanel = new ScalePanel();
    setScale(1.0);
    this.setLayout( new GridBagLayout() );
    this.scaleLabel.setHorizontalAlignment(JLabel.LEFT);
    this.add( scalePanel, new GridBagConstraints(0,0,2,1,1.0,1.0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0) );
    this.add( scaleLabel, new GridBagConstraints(0,1,1,1,0.5,0.0,GridBagConstraints.NORTHWEST,GridBagConstraints.NONE,new Insets(0,0,0,0),0,0) );
  }

  /**
   * Liefert die dargestellte Aufloesung in Pixel pro Meter.
   */
  public double getScaleInPixels() {
    return scalePanel.getScaleInPixels();
  }

  /**
   * Liefert die dargestellte Aufloesung in Metern pro Pixel.
   */
  public double getScaleInMeters() {
    return scalePanel.getScaleInMeters();
  }

  /**
   * Setzt die Aufloesung des Mass-Stabs und aktualisiert die Anzeige der
   * Komponente.
   * @param meters Meter, die durch einen Bildschirm-Pixel repraesentiert werden
   */
  public void setScale(double meters) {
    scalePanel.setScale(meters);
    scaleLabel.setText("1 pixel = "+numFormat.format(meters)+"m");
    repaint();
  }

  /**
   * Setzt die Aufloesung des Mass-Stabs und aktualisiert die Anzeige der
   * Komponente.
   * @param pixels Anzahl an Pixeln, die einen Meter repraesentieren werden
   * @exception Ill
   */
  public void setScale(long pixels) {
    scalePanel.setScale(pixels);
    scaleLabel.setText(pixels+" pixel = 1m");
    repaint();
  }
  
  /**
   * Nuetzlich wenn die Componente gedruckt (z.B. wenn ein Screenshot gemacht wird) wird. Dann werden wird der
   * Hintergrund auf WEISS gesetzt.
   *
   * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
   *         Kr&uuml;ger</a>
   */
  @Override
  public void print(Graphics g) {
      final Color orig = getBackground();
      setBackground(Color.WHITE);
      // remove the informatil   xxx pixel = 1m
      final String backupPixelText = scaleLabel.getText();
      scaleLabel.setText("");
      // wrap in try/finally so that we always restore the state
      try {
          super.print(g);
      } finally {
          setBackground(orig);
          scaleLabel.setText(backupPixelText);
      }
  }

}
