/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.styling;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Array;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.units.Unit;
import javax.xml.transform.TransformerException;

import org.apache.log4j.Logger;
import org.geotools.brewer.color.BrewerPalette;
import org.geotools.brewer.color.PaletteSuitability;
import org.geotools.brewer.color.SampleScheme;
import org.geotools.coverage.Category;
import org.geotools.coverage.GridSampleDimension;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.io.AbstractGridCoverage2DReader;
import org.geotools.data.FeatureSource;
import org.geotools.data.memory.MemoryFeatureCollection;
import org.geotools.factory.CommonFactoryFinder;
import org.geotools.factory.GeoTools;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureType;
import org.geotools.feature.GeometryAttributeType;
import org.geotools.filter.ConstantExpression;
import org.geotools.gui.swing.ExceptionMonitor;
import org.geotools.renderer.lite.RendererUtilities;
import org.geotools.resources.i18n.Vocabulary;
import org.geotools.resources.i18n.VocabularyKeys;
import org.geotools.styling.ColorMap;
import org.geotools.styling.ColorMapEntry;
import org.geotools.styling.ColorMapEntryImpl;
import org.geotools.styling.ColorMapImpl;
import org.geotools.styling.ExternalGraphic;
import org.geotools.styling.FeatureTypeStyle;
import org.geotools.styling.FeatureTypeStyleImpl;
import org.geotools.styling.Fill;
import org.geotools.styling.Graphic;
import org.geotools.styling.LineSymbolizer;
import org.geotools.styling.Mark;
import org.geotools.styling.PointSymbolizer;
import org.geotools.styling.PolygonSymbolizer;
import org.geotools.styling.RasterSymbolizer;
import org.geotools.styling.RasterSymbolizerImpl;
import org.geotools.styling.Rule;
import org.geotools.styling.RuleImpl;
import org.geotools.styling.SLDParser;
import org.geotools.styling.SLDTransformer;
import org.geotools.styling.Stroke;
import org.geotools.styling.Style;
import org.geotools.styling.StyleBuilder;
import org.geotools.styling.StyleFactory;
import org.geotools.styling.Symbolizer;
import org.geotools.styling.TextSymbolizer;
import org.geotools.util.NumberRange;
import org.jdom.Element;
import org.jdom.output.XMLOutputter;
import org.opengis.coverage.grid.Grid;
import org.opengis.coverage.grid.GridCoverage;
import org.opengis.filter.Filter;
import org.opengis.filter.FilterFactory2;
import org.opengis.filter.expression.Expression;
import org.opengis.filter.expression.Literal;
import org.opengis.filter.expression.NilExpression;

import schmitzm.geotools.FilterUtil;
import schmitzm.geotools.feature.FeatureUtil;
import schmitzm.geotools.feature.FeatureUtil.GeometryForm;
import schmitzm.geotools.grid.GridUtil;
import schmitzm.lang.LangUtil;
import skrueger.geotools.LegendIconFeatureRenderer;
import skrueger.i8n.Translation;

import com.sun.xml.bind.StringInputStream;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.MultiPoint;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Polygon;

/**
 * Diese Klasse enthaelt Hilfsfunktionen zum GeoTools-Styling
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons Kr&uuml;ger</a>
 * 
 * @version 1.1
 */
public class StylingUtil {
	private static final Logger LOGGER = Logger.getLogger(StylingUtil.class);
	/** Transparente Farbe */
	public static final Color TRANSPARENT_COLOR = new Color(0, 0, 0, 0);

	/**
	 * Standard-Instanz eines {@link StyleBuilder}.
	 * 
	 * @Deprecated Bitte nicht benutzen. Ich habe schlechte Efahrungen damit.
	 *             Zumindest uin 2.4.
	 */
	public static final StyleBuilder STYLE_BUILDER = new StyleBuilder();

	/** Standard-Instanz eines {@link SLDTransformer}. */
	public final static SLDTransformer SLDTRANSFORMER = new SLDTransformer();

	/**
	 * Standard-Instanz eines {@link StyleBuilder}.
	 * 
	 * @deprecated wurde ersetzt durch {@link #STYLE_BUILDER}
	 */
	public static final StyleBuilder builder = STYLE_BUILDER;

	/** Standard-Instanz einer {@link StyleFactory} */
	public static final StyleFactory STYLE_FACTORY = CommonFactoryFinder
			.getStyleFactory(GeoTools.getDefaultHints());

	/**
	 * Standard-Instanz einer {@link StyleFactory}
	 * 
	 * @deprecated wurde ersetzt durch {@link #STYLE_FACTORY}
	 */
	public static final StyleFactory styleFactory = STYLE_FACTORY;

	/**
	 * TODO SK The following constants are used to speed up
	 * {@link #createSelectionStyle(Object)}. I will "redo" them soon and also
	 * document them. This is still under construction. TODO SK
	 */

	static final Literal size0 = FeatureUtil.FILTER_FACTORY2.literal(13);
	static final Literal size1 = FeatureUtil.FILTER_FACTORY2.literal(9);
	static final Literal size2 = FeatureUtil.FILTER_FACTORY2.literal(7);
	static final Literal size3 = FeatureUtil.FILTER_FACTORY2.literal(3);
	static final Literal zeroLit = FeatureUtil.FILTER_FACTORY2.literal(0);

	// static final Stroke SELECTION_STROKE1 =
	// STYLE_FACTORY.createStroke(STYLE_BUILDER.colorExpression(Color.YELLOW),
	// size2);

	static final Stroke SELECTION_STROKE1 = STYLE_FACTORY.createStroke(
			STYLE_BUILDER.colorExpression(Color.YELLOW.brighter()),
			FeatureUtil.FILTER_FACTORY2.literal(4), FeatureUtil.FILTER_FACTORY2
					.literal(1.), FeatureUtil.FILTER_FACTORY.literal("round"),
			FeatureUtil.FILTER_FACTORY.literal("round"), new float[0],
			FeatureUtil.FILTER_FACTORY2.literal(0), null, null);

	static final Stroke SELECTION_STROKE2 = STYLE_FACTORY.createStroke(
			STYLE_BUILDER.colorExpression(Color.BLACK),
			FeatureUtil.FILTER_FACTORY2.literal(1.5),
			FeatureUtil.FILTER_FACTORY2.literal(1.), FeatureUtil.FILTER_FACTORY
					.literal("round"), FeatureUtil.FILTER_FACTORY
					.literal("round"), new float[] { 3f, 3f },
			FeatureUtil.FILTER_FACTORY2.literal(0), null, null);

	static final Stroke SELECTION_STROKE3 = STYLE_FACTORY.createStroke(
			STYLE_BUILDER.colorExpression(Color.WHITE),
			FeatureUtil.FILTER_FACTORY2.literal(1.5),
			FeatureUtil.FILTER_FACTORY2.literal(1.), FeatureUtil.FILTER_FACTORY
					.literal("round"), FeatureUtil.FILTER_FACTORY
					.literal("round"), new float[] { 3f, 3f },
			FeatureUtil.FILTER_FACTORY2.literal(3), null, null);

	static final Graphic SELECTION_GRAPHIC1 = STYLE_FACTORY.createGraphic(
			new ExternalGraphic[0], new Mark[] { STYLE_FACTORY.createMark(
					FeatureUtil.FILTER_FACTORY2.literal("square"), null,
					STYLE_FACTORY.createFill(STYLE_BUILDER
							.colorExpression(Color.WHITE)), size1, zeroLit) },
			new org.geotools.styling.Symbol[0], FeatureUtil.FILTER_FACTORY2
					.literal(1), size1, zeroLit);

	static final Graphic SELECTION_GRAPHIC2 = STYLE_FACTORY.createGraphic(
			new ExternalGraphic[0], new Mark[] { STYLE_FACTORY.createMark(
					FeatureUtil.FILTER_FACTORY2.literal("circle"), null,
					STYLE_FACTORY.createFill(STYLE_BUILDER
							.colorExpression(Color.BLACK)), size2, zeroLit) },
			new org.geotools.styling.Symbol[0], FeatureUtil.FILTER_FACTORY2
					.literal(1), size2, zeroLit);

	static final Graphic SELECTION_GRAPHIC3 = STYLE_FACTORY.createGraphic(
			new ExternalGraphic[0], new Mark[] { STYLE_FACTORY.createMark(
					FeatureUtil.FILTER_FACTORY2.literal("circle"), null,
					STYLE_FACTORY.createFill(STYLE_BUILDER
							.colorExpression(Color.YELLOW.brighter())), size3,
					zeroLit) }, new org.geotools.styling.Symbol[0],
			FeatureUtil.FILTER_FACTORY2.literal(1), size3, zeroLit);
	static final Fill SELECTION_FILL1 = STYLE_FACTORY.createFill(STYLE_BUILDER
			.colorExpression(Color.RED), null, FeatureUtil.FILTER_FACTORY2
			.literal(1), STYLE_FACTORY.createGraphic(new ExternalGraphic[0],
			new Mark[] { STYLE_FACTORY.createMark(FeatureUtil.FILTER_FACTORY2
					.literal("circle"), null, STYLE_FACTORY
					.createFill(STYLE_BUILDER.colorExpression(Color.RED)),
					size3, zeroLit) }, new org.geotools.styling.Symbol[0],
			FeatureUtil.FILTER_FACTORY2.literal(1), size3, zeroLit));

	// /**
	// * Up to three {@link Style styles} suitable to present selected {@link
	// Feature features} are cached here. (SK)
	// */
	// private static Map<GeometryForm,Style> selectionStyles = new
	// HashMap<GeometryForm, Style>();

	/**
	 * Erstellt einen Default-Style fuer ein Geo-Objekt.
	 * 
	 * @param object
	 *            {@link GridCoverage2D},
	 *            {@link org.geotools.coverage.grid.io.AbstractGridCoverage2DReader}
	 *            oder {@link FeatureCollection}
	 * @return {@code null} falls kein Style generiert werden kann
	 * 
	 */
	public static Style createDefaultStyle(Object object) {
		Style style = STYLE_FACTORY.createStyle(); // SK.. nicer default than
		// null !

		if (object instanceof GridCoverage2D
				|| object instanceof org.geotools.coverage.grid.io.AbstractGridCoverage2DReader)
			style = GridUtil.createDefaultStyle();

		if (object instanceof FeatureCollection)
			style = FeatureUtil.createDefaultStyle((FeatureCollection) object);

		// SK.cb
		if (object instanceof GeometryAttributeType)
			style = FeatureUtil
					.createDefaultStyle((GeometryAttributeType) object);

		if (object instanceof FeatureSource)
			style = FeatureUtil.createDefaultStyle(((FeatureSource) object)
					.getSchema().getDefaultGeometry());
		// SK.ce

		return style;
	}

	/**
	 * Erzeugt eine {@link Category} fuer die NoData-Werte transparent
	 * dargestellt werden.
	 * 
	 * @param geoValue
	 *            Geo-Wert, der NoData repraesentiert
	 */
	public static Category createNoDataCategory(double geoValue) {
		return createNoDataCategory(0, geoValue);
	}

	/**
	 * Erzeugt eine {@link Category} fuer die NoData-Werte transparent
	 * dargestellt werden.
	 * 
	 * @param value
	 *            Sample-Wert der Category
	 * @param geoValue
	 *            Geo-Wert, der NoData repraesentiert
	 */
	public static Category createNoDataCategory(int value, double geoValue) {
		return new Category(Vocabulary
				.formatInternational(VocabularyKeys.NODATA),
				new Color[] { new Color(0, 0, 0, 0) }, new NumberRange(value,
						value), new NumberRange(geoValue, geoValue));
	}

	/**
	 * Erzeugt eine {@link GridSampleDimension} aus einer {@link ColorMap}.
	 * 
	 * @param name
	 *            Name fuer die {@link GridSampleDimension}
	 * @param colorMap
	 *            fuer jeden Eintrag der {@link ColorMap} wird eine
	 *            {@link Category} erzeugt
	 * @param noDataValues
	 *            Werte fuer die zusaetzliche {@link Category Categorys}
	 *            erstellt werden, welche die transparent dargestellt werden
	 *            (wenn {@code null} wird keine zusaetzliche {@link Category}
	 *            erstellt)
	 * @param unit
	 *            Einheit der Werte in der {@link GridSampleDimension} (kann
	 *            {@code null} sein)
	 */
	public static GridSampleDimension createDiscreteGridSampleDimension(
			String name, ColorMap colorMap, double[] noDataValues, Unit unit) {
		if (colorMap.getColorMapEntries().length == 0)
			return null;

		SortedMap<Integer, Category> categories = new TreeMap<Integer, Category>();
		if (noDataValues == null)
			noDataValues = new double[0];

		// Create a (discrete) Category for each ColorMapEntry
		int colorMapMin = Integer.MAX_VALUE;
		int colorMapMax = Integer.MIN_VALUE;
		for (ColorMapEntry cme : colorMap.getColorMapEntries()) {
			String label = cme.getLabel();
			Color color = getColorFromColorMapEntry(cme);
			double value = getQuantityFromColorMapEntry(cme);
			int intValue = (int) Math.round(value);
			colorMapMin = Math.min(colorMapMin, intValue);
			colorMapMax = Math.max(colorMapMax, intValue);
			// if ( label == null || label.trim().equals("") )
			// label = String.valueOf(intValue);
			if (label == null)
				label = "";
			categories.put(intValue, new Category(label, new Color[] { color },
					new NumberRange(intValue, intValue), new NumberRange(
							intValue, intValue)));
		}

		// If NoData-Value is set, create an additional Category
		for (double noDataValue : noDataValues) {
			int value = categories.lastKey() + 10; // value not already used
			categories.put(value, createNoDataCategory(value, noDataValue));
			colorMapMin = Math.min(colorMapMin, (int) noDataValue);
			colorMapMax = Math.max(colorMapMax, (int) noDataValue);
		}

		// Declare "all" values smaller/greater the color map values
		// automatically as NoData
		// int colorMapRange = colorMapMax - colorMapMin;
		// double lowerStart = colorMapMin - 10000*colorMapRange;
		// double higherEnd = colorMapMax + 10000*colorMapRange;
		final int lowerStart = Integer.MIN_VALUE;
		final int higherEnd = Integer.MAX_VALUE;
		categories.put(colorMapMax + 10, new Category("_autoNoData1_",
				new Color[] { new Color(0, 0, 0, 0) }, new NumberRange(
						colorMapMax + 1, colorMapMax + 2), new NumberRange(
						colorMapMax + 1, higherEnd)));
		if (colorMapMin < colorMapMax)
			categories.put(colorMapMin - 10, new Category(
					"_autoNoData2_",
					new Color[] { new Color(0, 0, 0, 0) },
					// new NumberRange(colorMapMin-2, colorMapMin-1), //
					// logischer, aber klappt nicht!?
					new NumberRange(colorMapMax + 3, colorMapMax + 4),
					new NumberRange(lowerStart, colorMapMin - 1)));

		// Create the GridSampleDimension
		GridSampleDimension gsd = new GridSampleDimension(name, categories
				.values().toArray(new Category[0]), unit).geophysics(true);

		for (Category c : (List<Category>) gsd.getCategories())
			LOGGER.debug(c.toString());

		return gsd;
	}

	/**
	 * Liefert die Farbpalette zu einem Style, wenn der Style aus einem
	 * {@link RasterSymbolizer} erzeugt wurde.
	 * 
	 * @param style
	 *            Style Wenn <code>null</code>, dann wird style auf
	 *            {@link GridUtil}.createDefaultStyle() gesetzt
	 * @return <code>null</code> wenn der Style nicht auf einem
	 *         {@link RasterSymbolizer} basiert.
	 */
	public static ColorMap getColorMapFromStyle(Style style) {

		if (style == null)
			style = GridUtil.createDefaultStyle();

		FeatureTypeStyle[] fts = style.getFeatureTypeStyles();
		if (fts == null || fts.length == 0)
			return null;
		Rule[] rule = fts[0].getRules();
		if (rule == null || rule.length == 0)
			return null;
		Symbolizer[] symb = rule[0].getSymbolizers();
		if (symb == null || symb.length == 0
				|| !(symb[0] instanceof RasterSymbolizer))
			return null;
		return ((RasterSymbolizer) symb[0]).getColorMap();

	}

	/**
	 * Ermittelt fuer einen Wert den {@link ColorMapEntry} aus einer
	 * {@link ColorMap}.
	 * 
	 * @param value
	 *            Wert
	 * @return {@code null}, wenn in der {@link ColorMap} kein expliziter
	 *         Eintrag fuer den Wert hinterlegt ist
	 */
	public static ColorMapEntry findColorMapEntry(ColorMap colorMap,
			double value) {
		for (ColorMapEntry entry : colorMap.getColorMapEntries())
			if (value == getQuantityFromColorMapEntry(entry))
				return entry;
		return null;
	}

	/**
	 * Kopiert eine {@link ColorMap}.
	 * 
	 * @return <code>null</code> wenn die uebergebene ColorMap <code>null</code>
	 *         ist
	 */
	public static ColorMap cloneColorMap(ColorMap colorMap) {
		if (colorMap == null)
			return null;
		ColorMapEntry[] entry = colorMap.getColorMapEntries();
		ColorMap newColorMap = new ColorMapImpl();
		newColorMap.setType(colorMap.getType()); // Hat noch keinen Effekt in
		// 2.3.2, aber dennoch, SK
		for (int i = 0; i < entry.length; i++)
			newColorMap.addColorMapEntry(cloneColorMapEntry(entry[i]));
		return newColorMap;
	}

	/**
	 * Prueft, ob zwei {@link ColorMap}-Instanzen gleich sind.
	 * 
	 * @param cm1
	 *            eine Farbpalette
	 * @param cm2
	 *            eine andere Farbpalette
	 */
	public static boolean colorMapsEqual(ColorMap cm1, ColorMap cm2) {
		if (cm1 == null ^ cm2 == null)
			return false;
		if (cm1 == cm2)
			return true;

		ColorMapEntry[] cme1 = cm1.getColorMapEntries();
		ColorMapEntry[] cme2 = cm2.getColorMapEntries();
		if (cme1.length != cme2.length)
			return false;

		// Farbpaletten-Eintraege muessen gleich sein und in gleicher
		// Reihenfolge vorliegen!
		for (int i = 0; i < cme1.length; i++)
			if (!colorMapEntriesEqual(cme1[i], cme2[i]))
				return false;
		return true;
	}

	/**
	 * Prueft, ob zwei {@link ColorMapEntry}-Instanzen gleich sind. Dies ist der
	 * Fall, wenn ihnen der gleiche Wert, die gleiche Farbe, die gleiche
	 * Transparenz und das gleiche Label zugeordnet ist.
	 * 
	 * @param cme1
	 *            ein Farbpaletten-Eintrag
	 * @param cme2
	 *            ein anderer Farbpaletten-Eintrag
	 */
	public static boolean colorMapEntriesEqual(ColorMapEntry cme1,
			ColorMapEntry cme2) {
		if (cme1 == null ^ cme2 == null)
			return false;
		if (cme1 == cme2)
			return true;

		// Label auf Gleichheit testen
		if (cme1.getLabel() == null ^ cme2.getLabel() == null || cme1 != null
				&& !cme1.getLabel().equals(cme2.getLabel()))
			return false;
		// Farbe auf Gleichheit testen
		Color col1 = getColorFromColorMapEntry(cme1);
		Color col2 = getColorFromColorMapEntry(cme2);
		if (col1 == null ^ col2 == null || col1 != null && !col1.equals(col2))
			return false;
		// Wert auf Gleichheit testen
		Double quan1 = getQuantityFromColorMapEntry(cme1);
		Double quan2 = getQuantityFromColorMapEntry(cme2);
		if (quan1 == null ^ quan2 == null || quan1 != null
				&& !quan1.equals(quan2))
			return false;
		// Transparenz auf Gleichheit testen
		Double op1 = getOpacityFromColorMapEntry(cme1);
		Double op2 = getOpacityFromColorMapEntry(cme2);
		if (op1 == null ^ op2 == null || op1 != null && !op1.equals(op2))
			return false;

		return true;
	}

	/**
	 * Erzeugt einen {@link ColorMapEntry}.
	 * 
	 * @param label
	 *            Label fuer den Farbpaletten-Eintrag
	 * @param quantity
	 *            Wert fuer den Farbpaletten-Eintrag
	 * @param color
	 *            Farbe fuer den Farbpaletten-Eintrag
	 * @param opacity
	 *            Transparenz fuer die Farbe des Farbpaletten-Eintrag
	 */
	public static ColorMapEntry createColorMapEntry(String label,
			Double quantity, Color color, double opacity) {
		ColorMapEntry newEntry = new ColorMapEntryImpl();
		try {
			newEntry.setLabel(label);
			newEntry.setColor(STYLE_BUILDER.colorExpression(color));
			newEntry.setOpacity(STYLE_BUILDER.literalExpression(opacity));
			newEntry.setQuantity(STYLE_BUILDER.literalExpression(quantity));
		} catch (Exception err) {
		}
		return newEntry;
	}

	/**
	 * Kopiert einen {@link ColorMapEntry}.
	 * 
	 * @return <code>null</code> wenn der uebergebene ColorMapEntry
	 *         <code>null</code> ist
	 */
	public static ColorMapEntry cloneColorMapEntry(ColorMapEntry colorMapEntry) {
		if (colorMapEntry == null)
			return null;

		return createColorMapEntry(colorMapEntry.getLabel(),
				getQuantityFromColorMapEntry(colorMapEntry),
				getColorFromColorMapEntry(colorMapEntry),
				getOpacityFromColorMapEntry(colorMapEntry));
	}

	/**
	 * Liefert den Wert eines Farbpaletten-Eintrag.
	 * 
	 * @param expression
	 *            Expression, die einen String oder einen Double liefert
	 * @return {@code null}, wenn <code>null</code> uebergeben wird
	 */
	public static Double getQuantityFromExpression(Expression expression) {
		// gt2-2.3.4:
		// if ( expression == null || expression.getValue(null) == null )
		// return null
		// Object exprValue = evaluate.getValue(null);
		// gt2-2.3.4:
		if (expression == null || expression.evaluate(null) == null)
			return null;
		Object exprValue = expression.evaluate(null);

		return (exprValue instanceof String) ? Double
				.valueOf((String) exprValue) : ((Number) exprValue)
				.doubleValue();
	}

	/**
	 * Liefert den Wert eines Farbpaletten-Eintrag.
	 * 
	 * @param entry
	 *            Farbpaletten-Eintrag
	 * @return {@code null}, wenn <code>null</code> uebergeben wird
	 */
	public static Double getQuantityFromColorMapEntry(ColorMapEntry entry) {
		if (entry == null)
			return null;
		return getQuantityFromExpression(entry.getQuantity());
	}

	/**
	 * Setzt den Wert eines Farbpaletten-Eintrag.
	 * 
	 * @param entry
	 *            Farbpaletten-Eintrag
	 * @param quantity
	 *            neuer Wert
	 */
	public static void setQuantityForColorMapEntry(ColorMapEntry entry,
			double quantity) {
		if (entry != null)
			entry.setQuantity(STYLE_BUILDER.literalExpression(quantity));
	}

	/**
	 * Liefert die Transparenz eines Farbpaletten-Eintrag.
	 * 
	 * @param expression
	 *            Expression, die einen String oder einen Double liefert
	 * @return 1.0, wenn <code>null</code> uebergeben wird
	 */
	public static Double getOpacityFromExpression(Expression expression) {
		// gt2-2.3.4:
		// if ( expression == null || expression.getValue(null) == null )
		// return 1.0
		// Object exprValue = evaluate.getValue(null);
		// gt2-2.3.4:
		if (expression == null || expression.evaluate(null) == null)
			return 1.0;
		Object exprValue = expression.evaluate(null);

		return (exprValue instanceof String) ? Double
				.valueOf((String) exprValue) : ((Number) exprValue)
				.doubleValue();
	}

	/**
	 * Liefert die Transparenz eines Farbpaletten-Eintrag.
	 * 
	 * @param entry
	 *            Farbpaletten-Eintrag
	 * @return 1.0, wenn <code>null</code> uebergeben wird
	 */
	public static Double getOpacityFromColorMapEntry(ColorMapEntry entry) {
		if (entry == null)
			return 1.0;
		return getOpacityFromExpression(entry.getOpacity());
	}

	/**
	 * Setzt die Transparenz eines Farbpaletten-Eintrag.
	 * 
	 * @param entry
	 *            Farbpaletten-Eintrag
	 * @param opacity
	 *            Transparenzwert
	 */
	public static void setOpacityForColorMapEntry(ColorMapEntry entry,
			double opacity) {
		if (entry != null)
			entry.setOpacity(STYLE_BUILDER.literalExpression(opacity));
	}

	/**
	 * Liefert die Farbe eines Farbpaletten-Eintrag.
	 * 
	 * @param expression
	 *            Expression, die einen String liefert
	 */
	public static Color getColorFromExpression(Expression expression) {
		if (expression == null)
			return null;

		if (expression instanceof ConstantExpression) {
			ConstantExpression a = (ConstantExpression) expression;
			Object obj = a.getValue();
			if (obj instanceof Color) {
				return (Color) obj;
			}
		}

		return Color.decode(expression.toString());

		// Old way Martin did it:
		// return Color.decode((String) expression.evaluate(null));

	}

	/**
	 * Liefert die Farbe eines Farbpaletten-Eintrag.
	 * 
	 * @param entry
	 *            Farbpaletten-Eintrag
	 */
	public static Color getColorFromColorMapEntry(ColorMapEntry entry) {
		// if ( entry == null || entry.getColor() == null ||
		// entry.getColor().getValue(null) == null ) // gt2-2.3.4
		if (entry == null || entry.getColor() == null
				|| entry.getColor().evaluate(null) == null) // gt2-2-4-2
			return null;

		// // Transparenz
		// Double opacity = getOpacityFromColorMapEntry(entry);

		return getColorFromExpression(entry.getColor());
	}

	/**
	 * Setzt die Farbe eines Farbpaletten-Eintrag.
	 * 
	 * @param entry
	 *            Farbpaletten-Eintrag
	 * @param color
	 *            eine Farbe
	 */
	public static void setColorForColorMapEntry(ColorMapEntry entry, Color color) {
		if (entry != null)
			entry.setColor(STYLE_BUILDER.colorExpression(color));
	}

	/**
	 * Erzeugt einen {@link Style} aus einem {@linkplain Element JDOM-Element},
	 * des eine SLD-Definition enthaelt
	 * 
	 * @param element
	 *            Element mit SLD-Definition
	 */
	public static Style createStyleFromSLD(Element element) {
		String xmlDefinition = new XMLOutputter().outputString(element);
		Style[] style = loadSLD(new ByteArrayInputStream(xmlDefinition
				.getBytes()));
		return style == null || style.length == 0 ? null : style[0];
	}

	/**
	 * Creates a {@link Style} for a {@link Grid} layer from a {@link ColorMap}
	 * 
	 * @param colorMap
	 *            If null, then defaultStyle for Grid will be used
	 * @param name
	 *            The name to give to the Style. null will result in name=
	 *            {@link GridUtil#UNNAMED_RASTER_STYLE_NAME} or
	 *            GridUtil#DEFAULT_RASTER_STYLE_NAME} (if no colorMap is given).
	 * 
	 * @param title
	 *            The Title to give to the Style. null will result in title=
	 *            {@link GridUtil#UNTITLED_RASTER_STYLE_TITLE} or
	 *            {@link GridUtil#DEFAULT_RASTER_STYLE_TITLE} (if no colorMap is
	 *            given)
	 * 
	 * @return Always a {@link Style} that you can be applyed to a
	 *         {@link GridCoverage}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Style createStyleFromColorMap(ColorMap colorMap, String name,
			String title) {
		if (colorMap == null)
			return GridUtil.createDefaultStyle();

		if (name == null)
			// A colorMap was provided, but no name.
			name = GridUtil.UNNAMED_RASTER_STYLE_NAME;

		if (title == null)
			// A colorMap was provided, but no title.
			title = GridUtil.UNTITLED_RASTER_STYLE_TITLE;

		RasterSymbolizerImpl rasterSymbolizerImpl = new RasterSymbolizerImpl();

		// Entferne alle label informationen
		clearColorMapLabels(colorMap);

		rasterSymbolizerImpl.setColorMap(colorMap);
		Symbolizer[] symbolizers = { rasterSymbolizerImpl };
		RuleImpl ruleImpl = new RuleImpl(symbolizers) {
		};
		Rule[] rules = { ruleImpl };
		FeatureTypeStyleImpl featureTypeStyleImpl = new FeatureTypeStyleImpl(
				rules) {
		};

		Style style = STYLE_FACTORY.createStyle();
		style.setName(name);
		style.setTitle(title);
		style.addFeatureTypeStyle(featureTypeStyleImpl);
		return style;
	}

	/**
	 * Removes all label information from the {@link ColorMapEntry}s of the
	 * given {@link ColorMap}
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static ColorMap clearColorMapLabels(ColorMap colorMap) {
		for (ColorMapEntry cme : colorMap.getColorMapEntries()) {
			cme.setLabel(null);
		}
		return colorMap;
	}

	/**
	 * Creates a {@link Style} for a {@link Grid} layer from a {@link ColorMap}
	 * Title will be set to {@link GridUtil#UNTITLED_RASTER_STYLE_TITLE}
	 * 
	 * @param colorMap
	 *            If null, then defaultStyle for Grid will be used
	 * @param name
	 *            The name to give to the Style. null will result in name
	 *            {@link GridUtil#UNTITLED_RASTER_STYLE_TITLE}
	 * 
	 * @return Always a Style that you can apply to a {@link GridCoverage}.
	 *         Style has name {@link GridUtil#DEFAULT_RASTER_STYLE_NAME} if no
	 *         colormap was provided.
	 * 
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Style createStyleFromColorMap(ColorMap colorMap, String name) {
		return createStyleFromColorMap(colorMap, name, null);
	}

	/**
	 * Loads {@link Style}s from a SLD {@link InputStream}
	 * 
	 * @param url
	 *            {@link URL} to read the SLD from
	 * 
	 * @return An {@link Array} of {@link Style}s, can be length==0 if no
	 *         UserStyles in SLD file. null if file not exists
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Style[] loadSLD(URL url) {
		// LOGGER.debug("Loading styles from URL...");
		try {
			final InputStream openStream = url.openStream();
			return loadSLD(openStream);
		} catch (IOException e) {
			return null;
		}
	}

	/**
	 * Loads {@link Style}s from a SLD {@link InputStream}
	 * 
	 * @param inputStream
	 *            {@link InputStream} to read the SLD from
	 * 
	 * @return An {@link Array} of {@link Style}s, can be length==0. null if
	 *         file not exists
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Style[] loadSLD(InputStream inputStream) {
		final InputStream inputStream2 = inputStream;
		if (inputStream2 == null) {
			return null;
		}

		// LOGGER.debug("Loading styles from inputStream...");

		SLDParser stylereader;
		stylereader = new SLDParser(StylingUtil.STYLE_FACTORY);
		stylereader.setInput(inputStream);
		Style[] styles = stylereader.readXML();

		if (styles[0] == null) {
			LOGGER
					.warn(" ... no Styles recognized in inputStream. Will return empty styles[] array!");
		} else {
			// LOGGER.debug(" ... loaded " + styles.length
			// + " styles from inputStream, first style's name= "
			// + styles[0].getName());
		}

		// This is the main successful exit form loadSLD
		return styles;
	}
//
//	/**
//	 * Loads {@link Style}s from a SLD {@link InputStream} without setting a
//	 * {@link Charset}
//	 * 
//	 * @param inputStream
//	 *            {@link InputStream} to read the SLD from
//	 * 
//	 * @return An {@link Array} of {@link Style}s, can be length==0. null if
//	 *         file not exists
//	 * 
//	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
//	 *         Kr&uuml;ger</a>
//	 * 
//	 * @deprecated
//	 */
//	public static Style[] loadSLDOld(InputStream inputStream) {
//		final InputStream inputStream2 = inputStream;
//		if (inputStream2 == null) {
//			return null;
//		}
//
//		// Reader reader = new UTF8Reader(inputStream2);
//
//		LOGGER.debug("Loading styles from inputStream...");
//
//		SLDParser stylereader;
//		stylereader = new SLDParser(StylingUtil.STYLE_FACTORY, inputStream2);
//		Style[] styles = stylereader.readXML();
//
//		if (styles[0] == null) {
//			LOGGER
//					.warn(" ... no Styles recognized in inputStream. Will return empty styles[] array!");
//		} else {
//			LOGGER.debug(" ... loaded " + styles.length
//					+ " styles from inputStream, first style's name= "
//					+ styles[0].getName());
//		}
//
//		// This is the main successful exit form loadSLD
//		return styles;
//	}

	/**
	 * Loads {@link Style}s from a SLD {@link File}
	 * 
	 * @param sldFile
	 *            {@link File} to read the SLD from
	 * 
	 * @return An {@link Array} of {@link Style}s, can be length==0
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Style[] loadSLD(File sldFile) throws FileNotFoundException {
		LOGGER.debug("Loading styles from File...");

		final Style[] loadedSLD = loadSLD(new FileInputStream(sldFile));

		// TODO Maybe check if names and titles should be set to default
		// values...

		return loadedSLD;
	}

	/**
	 * Saves the {@link Style} to OGC SLD. Overwrites any existing file.
	 * 
	 * @param style
	 *            {@link Style} to save
	 * @param charset
	 *            The charset to use for the XML, e.g. <?xml version="1.0"
	 *            encoding="ISO-8859-1"?>. If null, ISO-8859-1 is used.
	 * 
	 * @throws TransformerException
	 * @throws IOException
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static final void saveStyleToSLD(Style style, File exportFile,
			Charset charset) throws TransformerException, IOException {

		if (charset != null) {
			LOGGER
					.warn("Explicitly setting the Chartset to "
							+ charset
							+ "! Steve thinks that this is a bad idea, because it doesn't"
							+ " transform the content but just declares the header.");
			SLDTRANSFORMER.setEncoding(charset);
		}
		SLDTRANSFORMER.setIndentation(2);
		SLDTRANSFORMER.transform(style, new FileWriter(exportFile));

		LOGGER.info("Saved a Style with charset "
				+ SLDTRANSFORMER.getEncoding().name() + " to " + exportFile);
	}

	/**
	 * Saves the {@link Style} to OGC SLD using ISO-8859-1 as charset.
	 * Overwrites any existing file.
	 * 
	 * @param style
	 *            {@link Style} to save
	 * 
	 * @throws TransformerException
	 * @throws IOException
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static final void saveStyleToSLD(Style style, File exportFile)
			throws TransformerException, IOException {
		saveStyleToSLD(style, exportFile, Charset.forName("ISO-8859-1"));
	}

	/**
	 * SLD Rules können die Paramter MinScaleDenominator und MaxScaleDenominator
	 * enthalten. Dadurch können Elemente für manche Zoom-Stufen deaktiviert
	 * werden.
	 * 
	 * Kommentar: "Sichtbarkeit" bezieht es nicht darauf, ob die Elemente auf
	 * dem Kartenausschnitt sichtbar sind, sondern um die SLD Regeln, welche das
	 * Feature evt. nur Scalenabhängig zeichnen.<br/>
	 * SK 19.6.2009:Bei einem {@link PolygonSymbolizer} zählt nicht der Rand,
	 * sondern nur ob eine Füllung vorhanden ist! Man kann dann zwar leider
	 * nicht auf ein Rand eines Polygons ohne Fill klicken, aber dafür kann man
	 * durch die ungefüllte Fläche klicken.
	 * 
	 * @param fc
	 *            Die zu filternde FeatureCollection. Diese wird nicht
	 *            verändert.
	 * @param style
	 *            Der Style, mit dem die Features gerendert werden (z.b.
	 *            layer.getStyle() )
	 * 
	 * @param scaleDenominator
	 *            Der aktuelle ScaleDenomitor für den die Sichtbarkeit ermittelt
	 *            wird.
	 * 
	 * @see RendererUtilities.calculateOGCScale
	 * 
	 * @return Eine FeatureCollection in welcher nur die Features enthalten
	 *         sind, welche bei aktuellen Scale mit dem übergebenen Style
	 *         gerendert werden.
	 * 
	 *         TODO Was ist mit raster?!
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static MemoryFeatureCollection filterSLDVisibleOnly(
			final FeatureCollection fc, final Style style,
			final Double scaleDenominator) {

		// Eine im Speicher gehaltene FeatureCollection der sichtbaren
		final MemoryFeatureCollection fcVisible = new MemoryFeatureCollection(
				fc.getFeatureType());

		// Prüfen aller Features in der Collection
		final Iterator<Feature> fcIt = fc.iterator();

		try {

			while (fcIt.hasNext()) {
				final Feature feature = fcIt.next();

				for (final FeatureTypeStyle fts : style.getFeatureTypeStyles()) {
					final List<Rule> rules = fts.rules();

					for (final Rule rule : rules) {
						final double maxScaleDenominator = rule
								.getMaxScaleDenominator();
						final double minScaleDenominator = rule
								.getMinScaleDenominator();

						// 1. Check: Trifft die Rule auf den aktuellen Scale der
						// JMapPane?
						if ((minScaleDenominator > scaleDenominator)
								|| (scaleDenominator > maxScaleDenominator)) {
							continue;
						}

						// 2. Check: Eine Rule ist gülrig, aber trifft der
						// Filter
						// für dieses Feature?
						final Filter rFilter = rule.getFilter();
						if ((rFilter != null) && (!rFilter.evaluate(feature)))
							continue;

						// 3. Check: Passt einer der Symbolizer dieser Rule zu
						// dem
						// Feature? Dieser Check wird in
						// "real world" meistens klappen, es sei denn, ein
						// Hansel
						// gibt einen LineSymbolizer für ein PointLayer an.
						boolean passt = false;
						for (final Symbolizer symb : rule.getSymbolizers()) {
							final Geometry geom = feature.getDefaultGeometry();
							if ((geom instanceof MultiPoint)
									|| (geom instanceof com.vividsolutions.jts.geom.Point)) {
								if (symb instanceof PointSymbolizer) {
									if (((PointSymbolizer) symb).getGraphic() != null)
										passt = true;
									break;
								}
							} else if ((geom instanceof MultiLineString)
									|| (geom instanceof LineString)) {
								if (symb instanceof LineSymbolizer) {
									if (((LineSymbolizer) symb).getStroke() != null)
										passt = true;
									break;
								}
							} else if ((geom instanceof MultiPolygon)
									|| (geom instanceof Polygon)) {
								if (symb instanceof PolygonSymbolizer) {
									if (((PolygonSymbolizer) symb).getFill() != null)
										passt = true;
									break;
								}
							} else {
								LOGGER
										.warn("Unbekannter geom Typ! Gehe von sichbar aus!");
								passt = true;
							}
						}
						if (!passt)
							continue;

						// Nun sollte das feature wiklich sichtbar sein
						fcVisible.add(feature);
					}
				}
			}
		} catch (final java.lang.IllegalStateException e) {
			LOGGER.error("Iterating over the features", e);
			/**
			 * SK: 14.Apri 2009. It happened a few time to that fcIt.hasNext
			 * suddenly threw an exception for the "africa countries.shp":
			 * Exception in thread "AWT-EventQueue-0"
			 * java.lang.IllegalStateException: ShapeType changed illegally from
			 * Polygon to Undefined at org.geotools.
			 * data.shapefile.shp.ShapefileReader.nextRecord(ShapefileReader
			 * .java:452) at org.geotools.data.shapefile.indexed.
			 * IndexedShapefileDataStore$Reader
			 * .next(IndexedShapefileDataStore.java:1272) at
			 * org.geotools.data.FIDFeatureReader.next(FIDFeatureReader.java:92)
			 * at org .geotools.data.FilteringFeatureReader.hasNext(
			 * FilteringFeatureReader .java:125) at
			 * org.geotools.data.store.FeatureReaderIterator.hasNext(
			 * FeatureReaderIterator.java:44) at
			 * schmitzm.geotools.feature.FeatureUtil
			 * .filterSLDVisibleOnly(FeatureUtil.java:216) *
			 * 
			 */
		} finally {
			fc.close(fcIt);
		}

		// LOGGER.info("filterSLDVisibleOnly removed "+ (fc.size() -
		// fcVisible.size()) + " features.");

		return fcVisible;
	}

	/**
	 * @param style
	 *            A {@link Style} to search for the first {@link TextSymbolizer}
	 *            that will be used for the given {@link Feature}.
	 * 
	 * @author Stefan A. Krüger
	 * 
	 * @return <code>null</code> or the first {@link TextSymbolizer} found in
	 *         the style that applies to the {@link Feature}.
	 */
	public static TextSymbolizer getTextSymbolizer(final Style style,
			final Feature f) {
		if (f == null)
			return null;
		try {
			if (style != null) {

				for (final FeatureTypeStyle fts : style.getFeatureTypeStyles()) {
					for (final Rule r : fts.getRules()) {

						final Filter filter = r.getFilter();
						//
						// System.out.println(f);
						// System.out.println(filter);
						// System.out.println();

						if (filter != null && (!filter.evaluate(f))) {
							continue;
						}

						for (final Symbolizer symb : r.getSymbolizers()) {
							if (symb instanceof TextSymbolizer) {
								return (TextSymbolizer) symb;
							}
						}

					}
				}

			}

		} catch (final Exception e) {
			LOGGER.error("error - filter API stuff?", e);
			return null;
		}
		return null;
	}

	/**
	 * @param style
	 *            A {@link Style} to search for all {@link TextSymbolizer}s . No
	 *            guarantee, that any one of them will ever be used for any
	 *            feature (think about filters).
	 * 
	 * @author Stefan A. Krüger
	 * 
	 * @return {@link List} or all {@link TextSymbolizer}s found in the
	 *         {@link Style}.
	 */
	public static List<TextSymbolizer> getTextSymbolizers(final Style style) {
		List<TextSymbolizer> results = new ArrayList<TextSymbolizer>();
		try {
			if (style != null) {

				for (final FeatureTypeStyle fts : style.getFeatureTypeStyles()) {
					for (final Rule r : fts.getRules()) {
						results.addAll(getTextSymbolizers(r.getSymbolizers()));
					}
				}

			}

		} catch (final Exception e) {
			LOGGER.error("error - filter API stuff?", e);
			return results;
		}
		return results;
	}

	/**
	 * @param symbolizers
	 *            List of Symbolizers to search for all {@link TextSymbolizer}s
	 *            . No guarantee, that any one of them will ever be used for any
	 *            feature (think about filters).
	 * 
	 * @author Stefan A. Krüger
	 * 
	 * @return {@link List} or all {@link TextSymbolizer}s found in the given
	 *         symbolizers.
	 */
	public static List<TextSymbolizer> getTextSymbolizers(
			Symbolizer[] symbolizers) {
		List<TextSymbolizer> results = new ArrayList<TextSymbolizer>();
		try {
			if (symbolizers != null) {

				for (final Symbolizer symb : symbolizers) {
					if (symb instanceof TextSymbolizer) {
						results.add((TextSymbolizer) symb);
					}
				}

			}

		} catch (final Exception e) {
			LOGGER.error("error - filter API stuff?", e);
			return results;
		}
		return results;
	}

	// /**
	// * Sorts a {@link ColorMap} by its quantity values. This is mentioned in
	// the
	// * OGC SLD Definition, page 52:
	// * <p>
	// * Quote: For example, a DEM raster giving elevations in meters above sea
	// * level can be translated to a colored image with a ColorMap. The
	// quantity
	// * attributes of a color-map are used for translating between numeric
	// * matrixes and color rasters and the ColorMap entries should be in order
	// of
	// * increasing numeric quantity so that intermediate numeric values can be
	// * matched to a color (or be interpolated between two colors). Labels may
	// be
	// * used for legends or may be used in the future to match character
	// values.
	// * Not all systems can support opacity in colormaps. The default opacity
	// is
	// * 1.0 (fully opaque). Defaults for quantity and label are
	// system-dependent.
	// * </p>
	// *
	// * @param colorMap
	// * {@link ColorMap} to sort by the quantity-field of the
	// * {@link ColorMapEntry}s
	// * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	// Kr&uuml;ger</a>
	// */
	// public static void sort(RemovableColorMapImpl colorMap) {
	// ColorMapEntry[] colorMapEntries = colorMap.getColorMapEntries();
	// SortedMap<Double, ColorMapEntry> tmap = new TreeMap<Double,
	// ColorMapEntry>();
	// // Add to a SortedMap
	// for (ColorMapEntry cme : colorMapEntries) {
	// tmap.put(getQuantityFromColorMapEntry(cme), cme);
	// colorMap.removeColorMapEntry(cme);
	// }
	// // Recreate ColorMap from sorted quantity values
	// for (ColorMapEntry cme : tmap.values()) {
	// colorMap.addColorMapEntry(cme);
	// }
	// }

	// /**
	// * Lets assume that every quantity value is maximally listed once in a
	// * {@link ColorMap}, then it makes some sense to use it as a key.
	// *
	// * @param quantityString
	// * quantitiy as String
	// * @return The first colorMap that has the same Quantity ( compared as
	// * Strings)
	// *
	// * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	// Kr&uuml;ger</a>
	// */
	// public static ColorMapEntry getColorMapEntryByQuantityString(
	// ColorMap colorMap, String quantityString) {
	// for (ColorMapEntry cme : colorMap.getColorMapEntries()) {
	// if (quantityString.equals(cme.getQuantity().toString()))
	// return cme;
	// }
	// return null;
	// }

	/**
	 * Clones a Style by converting it to XML and reading it back in.
	 * 
	 * @param style
	 *            the {@link Style} to be copied.
	 */
	public static Style clone(Style style) {

		String xml;
		try {
			xml = SLDTRANSFORMER.transform(style);

			SLDParser stylereader;
			stylereader = new SLDParser(StylingUtil.STYLE_FACTORY,
					new StringInputStream(xml));
			Style[] styles = stylereader.readXML();

			return styles[0];
		} catch (TransformerException e) {
			ExceptionMonitor.show(null, e);
			return null;
		}
	}

	public static Symbolizer clone(Symbolizer sym) {
		if (sym == null) {
			throw new RuntimeException("cant clone null");
		}

		/**
		 * Doing some corrections. Otherwise the Symbolizer can not be converted
		 * to XML
		 */
		if (sym instanceof PointSymbolizer) {
			PointSymbolizer ps = (PointSymbolizer) sym;
			Graphic graphic = ps.getGraphic();

			// That bugfix was needed in 2.4.5 before we patched the
			// StyleFactoryImpl
			if (graphic.getSize() instanceof NilExpression) {
				graphic.setSize(FilterUtil.FILTER_FAC.literal(10.));
			}

			try {
				SLDTRANSFORMER.transform(graphic);
			} catch (TransformerException e) {
				LOGGER.error("gr", e);
			}
		}

		Style style = STYLE_BUILDER.createStyle(sym); // TODO Mag ich das?

		Filter allwaysTrueFilter = FilterUtil.FILTER_FAC2.equals(
				FilterUtil.FILTER_FAC2.literal("1"), FilterUtil.FILTER_FAC2
						.literal("1"));

		style.getFeatureTypeStyles()[0].getRules()[0]
				.setFilter(allwaysTrueFilter);

		String xml;
		try {
			xml = SLDTRANSFORMER.transform(style);

			SLDParser stylereader;
			stylereader = new SLDParser(StylingUtil.STYLE_FACTORY,
					new StringInputStream(xml));
			Style[] styles = stylereader.readXML();

			return (Symbolizer) styles[0].getFeatureTypeStyles()[0].getRules()[0]
					.getSymbolizers()[0];
		} catch (TransformerException e) {
			ExceptionMonitor.show(null, e);
			return STYLE_FACTORY.createPointSymbolizer(); // TODO Abhängig vom
			// sym geometry type
		}

	}

	/**
	 * Clons a {@link Graphic} element
	 * 
	 * @param graphicFill
	 * @return
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Graphic clone(Graphic graphic) {
		try {
			PointSymbolizer ps = STYLE_FACTORY.createPointSymbolizer(graphic,
					"ACHTUNG");

			Style style = STYLE_BUILDER.createStyle(ps);

			final String xml = SLDTRANSFORMER.transform(style);

			SLDParser stylereader;
			stylereader = new SLDParser(StylingUtil.STYLE_FACTORY,
					new StringInputStream(xml));
			Style[] styles = stylereader.readXML();

			PointSymbolizer clonedPs = (PointSymbolizer) styles[0]
					.getFeatureTypeStyles()[0].getRules()[0].getSymbolizers()[0];

			Graphic clonedG = clonedPs.getGraphic();

			return clonedG;

		} catch (TransformerException e) {
			LOGGER.error("Cloning a Graphic", e);
		}
		return null;
	}

	/**
	 * Creates a new {@link BrewerPalette} with only one scheme (which includes
	 * all colors). The suitablity of the palette is set to "unknown" for all
	 * viewer types.
	 * 
	 * @param name
	 *            name for the palette (also the description is set to this
	 *            value)
	 * @param colors
	 *            colors for the palette
	 */
	public static BrewerPalette createBrewerPalette(String name, Color[] colors)
			throws IOException {
		BrewerPalette palette = new BrewerPalette();
		palette.setColors(colors);
		palette.setName(name);
		palette.setDescription(name);

		// set suitability to UNKNOWN for all viewers
		PaletteSuitability suitability = new PaletteSuitability();
		suitability.setSuitability(colors.length, new String[] { "?", "?", "?",
				"?", "?", "?" });
		palette.setPaletteSuitability(suitability);

		// set the only scheme for the palette (include all colors)
		int[] schema = new int[colors.length];
		for (int j = 0; j < schema.length; j++)
			schema[j] = j;
		SampleScheme sampleScheme = new SampleScheme();
		sampleScheme.setSampleScheme(schema.length, schema);
		palette.setColorScheme(sampleScheme);

		return palette;
	}

	/**
	 * @param geoObject
	 *            {@link GridCoverage2D}, {@link AbstractGridCoverage2DReader},
	 *            {@link FeatureCollection}, {@link GeometryAttributeType} or
	 *            {@link FeatureSource}
	 * 
	 * @return a {@link Style} that is suitable to identify features of the
	 *         given type as selected items.
	 */
	public static Style createSelectionStyle(Object geoObject) {

		if (geoObject instanceof GridCoverage2D
				|| geoObject instanceof org.geotools.coverage.grid.io.AbstractGridCoverage2DReader) {
			// Wenn irgendwann mal selection für raster möglich ist, dann hier
			// einen schöneren style erstellen.
			return GridUtil.createDefaultStyle();
		}

		// We have vector-data. Now let's determine the type...
		GeometryForm geometryForm = null;

		if (geoObject instanceof FeatureCollection) {
			geometryForm = FeatureUtil
					.getGeometryForm((FeatureCollection) geoObject);
		} else if (geoObject instanceof GeometryAttributeType) {
			geometryForm = FeatureUtil
					.getGeometryForm((GeometryAttributeType) geoObject);
		} else if (geoObject instanceof FeatureSource) {
			geometryForm = FeatureUtil
					.getGeometryForm((FeatureSource) geoObject);
		}

		// /**
		// * Let's see if we have a cached version of the requested style.
		// */
		// if (selectionStyles.containsKey(geometryForm))
		// return selectionStyles.get(geometryForm);

		Style style = STYLE_FACTORY.createStyle();
		style.addFeatureTypeStyle(STYLE_FACTORY.createFeatureTypeStyle());
		Symbolizer[] symbolizers = new Symbolizer[0];

		switch (geometryForm) {

		case POINT:
			PointSymbolizer ps = STYLE_FACTORY.createPointSymbolizer();
			ps.setGraphic(SELECTION_GRAPHIC1);
			symbolizers = LangUtil.extendArray(symbolizers, ps);

			PointSymbolizer ps2 = STYLE_FACTORY.createPointSymbolizer();
			ps2.setGraphic(SELECTION_GRAPHIC2);
			symbolizers = LangUtil.extendArray(symbolizers, ps2);

			PointSymbolizer ps3 = STYLE_FACTORY.createPointSymbolizer();
			ps3.setGraphic(SELECTION_GRAPHIC3);
			symbolizers = LangUtil.extendArray(symbolizers, ps3);

			break;
		case POLYGON:

			PolygonSymbolizer polS0 = STYLE_FACTORY.createPolygonSymbolizer();

			Fill SELECTION_FILL0 = STYLE_FACTORY.createFill(STYLE_BUILDER
					.colorExpression(Color.YELLOW.brighter()));

			polS0.setFill(SELECTION_FILL0);
			symbolizers = LangUtil.extendArray(symbolizers, polS0);

			PolygonSymbolizer polS = STYLE_FACTORY.createPolygonSymbolizer();
			polS.setFill(SELECTION_FILL1);
			symbolizers = LangUtil.extendArray(symbolizers, polS);

		case LINE:
			LineSymbolizer ls = STYLE_FACTORY.createLineSymbolizer();
			ls.setStroke(SELECTION_STROKE1);
			symbolizers = LangUtil.extendArray(symbolizers, ls);

			LineSymbolizer ls2 = STYLE_FACTORY.createLineSymbolizer();
			ls2.setStroke(SELECTION_STROKE2);
			symbolizers = LangUtil.extendArray(symbolizers, ls2);

			LineSymbolizer ls3 = STYLE_FACTORY.createLineSymbolizer();
			ls3.setStroke(SELECTION_STROKE3);
			symbolizers = LangUtil.extendArray(symbolizers, ls3);
			break;

		default:
			throw new IllegalArgumentException("Provide a suitable object.");
		}

		style.getFeatureTypeStyles()[0].setRules(new Rule[] { STYLE_FACTORY
				.createRule() });
		style.getFeatureTypeStyles()[0].getRules()[0]
				.setSymbolizers(symbolizers);

		// selectionStyles.put(geometryForm, style);

		return style;

	}

	public static Style createStyleSimple(Object geoObject, Color color,
			Color color2) {
		if (geoObject instanceof GridCoverage2D
				|| geoObject instanceof org.geotools.coverage.grid.io.AbstractGridCoverage2DReader) {
			// Wenn irgendwann mal selection für raster möglich ist, dann hier
			// einen schöneren style erstellen.
			return GridUtil.createDefaultStyle();
		}

		// We have vector-data. Now let's determine the type...
		GeometryForm geometryForm = null;

		if (geoObject instanceof FeatureCollection) {
			geometryForm = FeatureUtil
					.getGeometryForm((FeatureCollection) geoObject);
		} else if (geoObject instanceof GeometryAttributeType) {
			geometryForm = FeatureUtil
					.getGeometryForm((GeometryAttributeType) geoObject);
		} else if (geoObject instanceof FeatureSource) {
			geometryForm = FeatureUtil
					.getGeometryForm((FeatureSource) geoObject);
		}

		// /**
		// * Let's see if we have a cached version of the requested style.
		// */
		// if (selectionStyles.containsKey(geometryForm))
		// return selectionStyles.get(geometryForm);

		Style style = STYLE_FACTORY.createStyle();
		style.addFeatureTypeStyle(STYLE_FACTORY.createFeatureTypeStyle());
		Symbolizer[] symbolizers = new Symbolizer[0];

		switch (geometryForm) {

		case POINT:
			PointSymbolizer ps = STYLE_BUILDER
					.createPointSymbolizer(STYLE_BUILDER.createGraphic(null,
							STYLE_BUILDER.createMark("circle"), null));
			ps.getGraphic().getMarks()[0].setFill(STYLE_BUILDER.createFill(
					color, 0.5));
			ps.getGraphic().getMarks()[0].setSize(STYLE_BUILDER
					.literalExpression(8.));
			ps.getGraphic().setSize(STYLE_BUILDER.literalExpression(8.));
			ps.getGraphic().getMarks()[0].setStroke(STYLE_BUILDER
					.createStroke(color2));
			symbolizers = LangUtil.extendArray(symbolizers, ps);

			break;
		case POLYGON:
			PolygonSymbolizer polS = STYLE_BUILDER.createPolygonSymbolizer(
					STYLE_BUILDER.createStroke(color2), STYLE_BUILDER
							.createFill(color));
			symbolizers = LangUtil.extendArray(symbolizers, polS);

		case LINE:
			LineSymbolizer ls = STYLE_BUILDER
					.createLineSymbolizer(STYLE_BUILDER.createStroke(color));
			symbolizers = LangUtil.extendArray(symbolizers, ls);
			break;

		default:
			throw new IllegalArgumentException("Provide a suitable object.");
		}

		style.getFeatureTypeStyles()[0].setRules(new Rule[] { STYLE_FACTORY
				.createRule() });
		style.getFeatureTypeStyles()[0].getRules()[0]
				.setSymbolizers(symbolizers);

		return style;
	}

	/**
	 * Creates a {@link Box} that shows a legend for a list of
	 * {@link FeatureTypeStyle}s and a targeted featureType
	 * 
	 * @param featureType
	 *            If this a legend for Point, Polygon or Line?
	 * @param featureTypeStyles
	 *            The Styles to presen in this legend
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Box createLegendPanel(FeatureTypeStyle[] featureTypeStyles,
			FeatureType featureType) {

		Box box = new Box(BoxLayout.Y_AXIS) {

			/**
			 * Nuetzlich wenn die Componente gedruckt (z.B. wenn ein Screenshot
			 * gemacht wird) wird. Dann werden wird der Hintergrund auf WEISS
			 * gesetzt.
			 * 
			 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
			 *         Kr&uuml;ger</a>
			 */
			@Override
			public void print(Graphics g) {
				final Color orig = getBackground();
				setBackground(Color.WHITE);
				// wrap in try/finally so that we always restore the state
				try {
					super.print(g);
				} finally {
					setBackground(orig);
				}
			}
		};

		for (FeatureTypeStyle ftStyle : featureTypeStyles) {

			// One child-node for every rule
			Rule[] rules = ftStyle.getRules();
			for (Rule rule : rules) {

				/**
				 * Let's not create a hbox for Rules that only contain
				 * TextSymbolizers
				 */
				if (StylingUtil.getTextSymbolizers(rule.getSymbolizers())
						.size() == rule.getSymbolizers().length)
					continue;

				Box hbox = new Box(BoxLayout.X_AXIS) {

					/**
					 * Nuetzlich wenn die Componente gedruckt (z.B. wenn ein
					 * Screenshot gemacht wird) wird. Dann werden wird der
					 * Hintergrund auf WEISS gesetzt.
					 * 
					 * @author <a href="mailto:skpublic@wikisquare.de">Stefan
					 *         Alfons Kr&uuml;ger</a>
					 */
					@Override
					public void print(Graphics g) {
						final Color orig = getBackground();
						setBackground(Color.WHITE);
						// wrap in try/finally so that we always restore the
						// state
						try {
							super.print(g);
						} finally {
							setBackground(orig);
						}
					}
				};

				/**
				 * The size of the legend Symbol is dependent on the size of the
				 * font.
				 */
				final int fontHeight = new JLabel().getFontMetrics(
						new JLabel().getFont()).getHeight();
				final Dimension ICON_SIZE = new Dimension(23,
						fontHeight > 5 ? fontHeight : 5);

				// ****************************************************************************
				// Create the actual icon
				// ****************************************************************************
				final BufferedImage imageForRule = LegendIconFeatureRenderer
						.getInstance().createImageForRule(rule, featureType,
								ICON_SIZE);

				// LOGGER.debug("Creating a new Legend Image for RUle name =
				// "+rule.getName());

				ImageIcon legendIcon = new ImageIcon(imageForRule);

				final JLabel iconLabel = new JLabel(legendIcon);
				hbox.setAlignmentX(0f);
				hbox.add(iconLabel);
				hbox.add(Box.createHorizontalStrut(3));

				// ****************************************************************************
				// The labeltext is read from the SLD. Its either the title
				// directly, or interpreted as OneLine Code
				// ****************************************************************************
				final String rawText = rule.getTitle();

				Translation labelT = new Translation();
				labelT.fromOneLine(rawText);

				final JLabel classTitleLabel = new JLabel(labelT.toString());
				hbox.add(classTitleLabel);
				classTitleLabel.setLabelFor(iconLabel); // Whatever it is good
				// for

				box.add(hbox);

			}
		}

		return box;
	}

	//
	// /**
	// * Doubles the number of palettes by adding the reverse palette to every
	// * palette. The name of the reverse palette is the name of the original
	// * palette plus the letter R.
	// *
	// * @author SK
	// * @param palettes The original list of {@link BrewerPalette
	// BrewerPalettes}
	// */
	// public static BrewerPalette[] addReversePalettes(BrewerPalette[]
	// palettes) {
	// BrewerPalette[] newPalettes = new BrewerPalette[palettes.length * 2];
	//
	// int pos = 0;
	// for (BrewerPalette bp : palettes) {
	// newPalettes[pos] = bp;
	// pos++;
	// Color[] newColors = new Color[bp.getMaxColors()];
	//
	// List<Color> colorList =
	// java.util.Arrays.asList(bp.getColors(bp.getMaxColors()));
	// Collections.reverse(colorList);
	// colorList.toArray(newColors);
	//
	// try {
	// newPalettes[pos] = StylingUtil.createBrewerPalette(bp.getName()
	// + "R", newColors);
	// newPalettes[pos].setPaletteSuitability(bp
	// .getPaletteSuitability());
	// newPalettes[pos].setDescription(bp.getDescription()+" reversed");
	// newPalettes[pos].setType(bp.getType());
	// newPalettes[pos].setColorScheme(bp.getColorScheme());
	// } catch (IOException e) {
	// LOGGER.error("",e);
	// newPalettes[pos] = bp;
	// }
	//
	// pos++;
	// }
	//
	// return newPalettes;
	// }

	/**
	 * Replaces the "main" color in a given {@link Symbolizer} element
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void replaceSymbolizerColor(Symbolizer symb, Color oldColor,
			Color newColor) {
		if (symb == null)
			return;

		if (symb instanceof PointSymbolizer) {
			PointSymbolizer ps = (PointSymbolizer) symb;
			replacePointSymbolizerColor(ps, oldColor, newColor);
		}

		if (symb instanceof PolygonSymbolizer) {
			PolygonSymbolizer ps = (PolygonSymbolizer) symb;
			replacePolygonSymbolizerColor(ps, oldColor, newColor);
		}

		if (symb instanceof LineSymbolizer) {
			LineSymbolizer ps = (LineSymbolizer) symb;
			replaceLineSymbolizerColor(ps, oldColor, newColor);
		}

	}

	/**
	 * Replaces the "main" color in a given {@link LineSymbolizer} element
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void replaceLineSymbolizerColor(LineSymbolizer ps,
			Color oldColor, Color newColor) {
		if (ps == null)
			return;

		replaceStrokeColor(ps.getStroke(), oldColor, newColor);
	}

	/**
	 * Replaces the "main" color in a given {@link PointSymbolizer} element
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void replacePointSymbolizerColor(PointSymbolizer ps,
			Color oldColor, Color newColor) {
		if (ps == null)
			return;

		replaceGraphicColor(ps.getGraphic(), oldColor, newColor);
	}

	/**
	 * Replaces the "main" color in a given {@link PolygonSymbolizer} element
	 * 
	 * @param oldColor
	 * @param newColor
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 * @param ps
	 */
	public static void replacePolygonSymbolizerColor(PolygonSymbolizer ps,
			Color oldColor, Color newColor) {
		replaceFillColor(ps.getFill(), oldColor, newColor);
		replaceStrokeColor(ps.getStroke(), oldColor, newColor);
	}

	/**
	 * Replaces the "main" color in a given {@link Stroke} element
	 * 
	 * @param stroke
	 * @param oldColor
	 * @param newColor
	 */
	public static void replaceStrokeColor(Stroke stroke, Color oldColor,
			Color newColor) {
		if (stroke == null)
			return;

		if ((stroke.getColor() != null)
				&& (StylingUtil.getColorFromExpression(stroke.getColor())
						.equals(oldColor)))
			stroke
					.setColor(StylingUtil.STYLE_BUILDER
							.colorExpression(newColor));

		replaceGraphicColor(stroke.getGraphicFill(), oldColor, newColor);

		replaceGraphicColor(stroke.getGraphicStroke(), oldColor, newColor);

	}

	/**
	 * Replaces the "main" color in a given {@link Fill} element
	 * 
	 * @param fill
	 * @param oldColor
	 * @param newColor
	 */
	public static void replaceFillColor(Fill fill, Color oldColor,
			Color newColor) {
		if (fill == null)
			return;

		if ((fill.getColor() != null)
				&& (getColorFromExpression(fill.getColor()).equals(oldColor)))
			fill.setColor(STYLE_BUILDER.colorExpression(newColor));

		replaceGraphicColor(fill.getGraphicFill(), oldColor, newColor);

		if ((fill.getBackgroundColor() != null)
				&& (getColorFromExpression(fill.getBackgroundColor())
						.equals(oldColor)))
			fill.setBackgroundColor(STYLE_BUILDER.colorExpression(newColor));
	}

	/**
	 * Replaces the "main" color in a given {@link Graphic} element
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void replaceGraphicColor(Graphic graphic, Color oldColor,
			Color newColor) {
		if (graphic == null)
			return;

		if ((graphic.getMarks() != null) && (graphic.getMarks().length > 0)) {

			// Checking for Colors in Marks...
			for (Mark m : graphic.getMarks()) {

				replaceFillColor(m.getFill(), oldColor, newColor);

				replaceStrokeColor(m.getStroke(), oldColor, newColor);

			}
		}
	}

	/**
	 * @return the {@link Color} used in the {@link Graphic} or null, if an
	 *         {@link ExternalGraphic} is used.
	 * 
	 * @param graphic
	 *            If <code>null</code> returns <code>null</code>
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getGraphicColor(Graphic graphic) {
		if (graphic == null)
			return null;

		Color foundColor = null;

		if ((graphic.getMarks() != null) && (graphic.getMarks().length > 0)) {

			// Checking for Colors in Marks...
			for (Mark m : graphic.getMarks()) {

				if (foundColor == null)
					foundColor = getFillColor(m.getFill());

				if (foundColor == null)
					foundColor = getStrokeColor(m.getStroke());
			}

		}

		return foundColor;
	}

	/**
	 * @return the first {@link Color} used in a {@link Stroke}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getStrokeColor(Stroke stroke) {
		if (stroke == null)
			return null;

		if (stroke.getColor() != null)
			return StylingUtil.getColorFromExpression(stroke.getColor());

		if (getGraphicColor(stroke.getGraphicStroke()) != null)
			return getGraphicColor(stroke.getGraphicStroke());

		if (getGraphicColor(stroke.getGraphicFill()) != null)
			return getGraphicColor(stroke.getGraphicFill());

		return null;
	}

	/**
	 * @return the first {@link Color} used in a {@link Fill}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getFillColor(Fill fill) {
		if (fill == null)
			return null;

		Color foundColor = null;

		if (foundColor == null)
			foundColor = getColorFromExpression(fill.getColor());

		if (foundColor == null)
			foundColor = getGraphicColor(fill.getGraphicFill());

		if (foundColor == null)
			foundColor = getColorFromExpression(fill.getBackgroundColor());

		return foundColor;
	}

	/**
	 * @return the first {@link Color} used in a {@link Symbolizer}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getSymbolizerColor(Symbolizer symb) {
		if (symb == null)
			return null;

		if (symb instanceof PointSymbolizer) {
			PointSymbolizer ps = (PointSymbolizer) symb;
			return getPointSymbolizerColor(ps);
		}

		if (symb instanceof PolygonSymbolizer) {
			PolygonSymbolizer ps = (PolygonSymbolizer) symb;
			return getPolygonSymbolizerColor(ps);
		}

		if (symb instanceof LineSymbolizer) {
			LineSymbolizer ps = (LineSymbolizer) symb;
			return getLineSymbolizerColor(ps);
		}

		return null;
	}

	/**
	 * @return the first {@link Color} used in a {@link LineSymbolizer}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getLineSymbolizerColor(LineSymbolizer ps) {
		if (ps == null)
			return null;

		Color foundColor = null;

		if (foundColor == null)
			foundColor = getStrokeColor(ps.getStroke());

		return foundColor;
	}

	/**
	 * @return the first {@link Color} used in a {@link PolygonSymbolizer}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getPolygonSymbolizerColor(PolygonSymbolizer ps) {
		if (ps == null)
			return null;

		Color foundColor = null;

		if (foundColor == null)
			foundColor = getFillColor(ps.getFill());

		if (foundColor == null)
			foundColor = getStrokeColor(ps.getStroke());

		return foundColor;
	}

	/**
	 * @return the first {@link Color} used in a {@link PointSymbolizer}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getPointSymbolizerColor(PointSymbolizer ps) {
		if (ps == null)
			return null;

		return getGraphicColor(ps.getGraphic());
	}

	/***************************************************************************
	 * Copies all Values from one {@link TextSymbolizer} to another
	 * {@link TextSymbolizer}
	 * 
	 * @param from
	 *            {@link TextSymbolizer} source
	 * @param to
	 *            {@link TextSymbolizer} target. May not be <code>null</code>.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void copyAllValues(final TextSymbolizer from,
			final TextSymbolizer to) {
		to.setLabel(from.getLabel());

		final FilterFactory2 ff2 = FeatureUtil.FILTER_FACTORY2;

		to.getFonts()[0].setFontFamily(ff2.literal(from.getFonts()[0]
				.getFontFamily().toString()));
		to.getFonts()[0].setFontSize(ff2.literal(from.getFonts()[0]
				.getFontSize().toString()));
		to.getFonts()[0].setFontWeight(ff2.literal(from.getFonts()[0]
				.getFontWeight().toString()));
		to.getFonts()[0].setFontStyle(ff2.literal(from.getFonts()[0]
				.getFontStyle().toString()));
	}

}
